import numpy as np
from scipy import signal as si
from scipy.optimize import curve_fit

class TransferFunctionFit:
    def __init__(self, num, den, delay):
        self.num   = num   # transfer function numerator:   list
        self.den   = den   # transfer function denominator: list
        self.delay = delay # input-output delay:            int (seconds)
        self.R2    = None
        
    def gen_response(self, u, y0):
        # sampling interval in seconds
        dt = np.median((u.index[1:]-u.index[:-1]).seconds)
        
        # Delay input
        u_shift = np.repeat(u[0], int(round(self.delay/dt))+1)
        u       = np.array((u_shift.tolist() + u.values.tolist())\
                           [:u.shape[0]])-u[0]
            
        # Discrete time intervals            
        t  = np.arange(0,u.shape[0]*dt, dt)
        
        # Generate output   
        return si.TransferFunction(self.num,self.den).output(u,t)[1]+y0
    
    def optimize(self, u, y, iterations=1):
        for _ in range(iterations):
            
            # Set intial parameters and optimization bounds
            #------------------------------------------------------------------
            params_0 = self.num + self.den
            l_bounds = []
            u_bounds = []
            for el in params_0:
                if   el==0:l_bounds.append(-1e-100); u_bounds.append(1e-100 )  
                elif el>0 :l_bounds.append(el*0.01); u_bounds.append(el*1000)
                else      :l_bounds.append(el*1000.);u_bounds.append(el*.01 )
            if round(params_0[-1],1)==1: 
                l_bounds[-1]=1
                u_bounds[-1]=1+1e-15 
            if round(params_0[-2],2)==1 and round(params_0[-1],1)==0:
                l_bounds[-2]=1
                u_bounds[-2]=1+1e-15 
            
            # Fit numerator and denominator
            #------------------------------------------------------------------
            def model(u, *params):
                self.num   = params[:len(self.num)]
                self.den   = params[len(self.num):]
                return self.gen_response(u, y[0])
            params = curve_fit(model, u, y, 
                               p0 = params_0, 
                               bounds=(l_bounds,u_bounds))[0]
            self.num = params[:len(self.num)].tolist()
            self.den = params[len(self.num):].tolist()

            # Calc Delay
            #------------------------------------------------------------------
            if self.delay != 0:
                dt = np.median((u.index[1:]-u.index[:-1]).seconds)
                c_m = {}
                for de in range(0,int(u.shape[0]*0.2*dt),int(dt)):
                    self.delay = de
                    y_m        = self.gen_response(u, y[0])
                    c_m[de]    = np.corrcoef(y_m,y)[0,1]
                self.delay = max(c_m, key=c_m.get)
            
        #----------------------------------------------------------------------
        SSres   =  np.sum((y - self.gen_response(u, y[0]))**2)
        SStot   =  np.sum((y - np.mean(y))**2)
        self.R2 = (1 - SSres/SStot)

if __name__ == "__main__":
    import pandas as pd
    import datetime
    import matplotlib.pyplot as plt
    import os
    import sys
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from models.fopdt import first_order_model
    from models.ramp import integrator_model

    '''------------------------------------------------------------------------
       Compare Transfer function with base dy/dt models
    ----------------------------------------------------------------------- '''
    delay  = 0
    Kp     = 1
    Tc     = 10
    K      = 1    
    y0     = 0
    dt     = 5  # seconds
    period = 60 # seconds
    u0     = 0
    step   = 1
    
    start  = datetime.datetime(2021,1,1,0)
    end    = start + datetime.timedelta(seconds=period)
    t = np.arange(start,end,datetime.timedelta(seconds=dt)).astype(datetime.datetime)
    u = pd.Series(np.repeat(u0, t.shape[0]), index = t)
    u[1:] = u0+step
    
    fig, ax = plt.subplots(2,1, figsize=(12,15))
    # =========================================================================
    tff_  = TransferFunctionFit([Kp],[Tc,1],delay)
    fopd_ = first_order_model(Kp, Tc, delay, dt, y0, u0)
    fopd_ = [fopd_.gen_response(ui) for ui in u]
    fopd_.insert(0, y0)
    result = pd.DataFrame(np.c_[[u.index,u,tff_.gen_response(u,y0),fopd_[:-1]]].T)
    result.columns = ["Date", "u", "transferFunction", "dy/dt"]
    result.index = result.Date; result = result.drop(["Date"],axis=1) 
    result.index = np.arange(-dt,dt*(result.shape[0]-1),dt)
    result.plot(drawstyle="steps-post",ax=ax[0])
    for i,j in zip(result.index,result["transferFunction"]):
        ax[0].annotate(str(round(j,2)),xy=(i,j),ha="right",va="center")
    for i,j in zip(result.index,result["dy/dt"]):
        ax[0].annotate(str(round(j,2)),xy=(i,j),ha="right",va="center")
    ax[0].grid(True)
    # =========================================================================
    tff_  = TransferFunctionFit([K],[1,0],delay)
    foim_ = integrator_model(K, delay, dt, y0, 0)
    foim_ = [foim_.gen_response(ui) for ui in u]
    foim_.insert(0, y0)
    result = pd.DataFrame(np.c_[[u.index,u,tff_.gen_response(u,y0),foim_[:-1]]].T)
    result.columns = ["Date", "u", "transferFunction", "dy/dt"]
    result.index = result.Date; result = result.drop(["Date"],axis=1)
    result.index = np.arange(-dt,dt*(result.shape[0]-1),dt)
    result.plot(drawstyle="steps-post",ax=ax[1])
    for i,j in zip(result.index,result["transferFunction"]):
        ax[1].annotate(str(round(j,2)),xy=(i,j),ha="right",va="center")
    for i,j in zip(result.index,result["dy/dt"]):
        ax[1].annotate(str(round(j,2)),xy=(i,j),ha="right",va="center")
    ax[1].grid(True)
    plt.show()


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def multi_yax_plot(ax, 
                   leftax_df=pd.DataFrame(),  # left  axis data frame 
                   rightax_df=pd.DataFrame(), # right axis data frame
                   leftax_limits=None,        # options include:
                   rightax_limits=None,       # None | [low,high] | [[low,high], [low,high], [low,high]]
                   leftax_style="-",          # options include:
                   rightax_style="-",         # "-", ".", ".-" | ["-",".",":"]
                   colors="bright",           # "palette" | ["k", "r", "g"]
                   linewidth=None,
                   markersize=None,
                   ylabel_fontsize=15):
    
    def set_spines(ax,which="left",loc=0,color="k"):
        if which!= "host":
            left_right = ["right","left"]
            if which == "left": left_right = ["left","right"]
            ax.spines[left_right[0]].set_position(("axes", loc))
            ax.spines[left_right[0]].set_visible(True)
            ax.spines[left_right[0]].set_linewidth(3)
            ax.spines[left_right[1]].set_visible(False)
            ax.spines["bottom"].     set_visible(False)
            ax.spines["top"].        set_visible(False)
            plt.setp(ax.spines.values(), color=color)
            ax.yaxis.set_label_position(which)
            ax.yaxis.set_ticks_position(which)
            ax.yaxis.label.set_color(color)
            ax.tick_params(axis='y', colors=color,width=2,size=5)
        else:
            ax.axes.get_yaxis().set_visible(False)
            ax.spines["right"].set_visible(False)
            ax.spines["left"].set_visible(False)
            ax.spines["top"].set_visible(False)
            ax.spines["bottom"].set_visible(True)
    
    if leftax_limits  is None:leftax_limits =[[None,None] for _ in range(leftax_df. shape[1])]
    if rightax_limits is None:rightax_limits=[[None,None] for _ in range(rightax_df.shape[1])]
    if not any(isinstance(x, list) for x in leftax_limits): leftax_limits =[[leftax_limits[0], leftax_limits[1]] for _ in range(leftax_df. shape[1])]
    if not any(isinstance(x, list) for x in rightax_limits):rightax_limits=[[rightax_limits[0],rightax_limits[1]]for _ in range(rightax_df.shape[1])]
    if not isinstance(leftax_style, list): leftax_style =[leftax_style  for _ in range(leftax_df. shape[1])]
    if not isinstance(rightax_style,list): rightax_style=[rightax_style for _ in range(rightax_df.shape[1])]
    if not isinstance(colors,list): colors = sns.color_palette(colors, leftax_df.shape[1]+rightax_df.shape[1])
    
    set_spines(ax, which="host")
    ax_left  = [ax.twinx() for _ in range(leftax_df. shape[1])]
    ax_right = [ax.twinx() for _ in range(rightax_df.shape[1])]
    
    for i,col in enumerate(leftax_df.columns):
        set_spines(ax_left[i], which="left", loc=-(i*1.2)/plt.gcf().get_size_inches()[0], color=colors[i])
        ax_left[i].set_ylabel(col,fontweight='bold',fontsize=ylabel_fontsize,loc="bottom")
        ax_left[i].plot(leftax_df.index,leftax_df[col],
                        leftax_style[i],color=colors[i],
                        linewidth=linewidth,markersize=markersize,
                        drawstyle="steps-post")
        ax_left[i].set_ylim(leftax_limits[i][0],leftax_limits[i][1])

    if leftax_df.columns.shape[0] == 0: i=0
    for j,col in enumerate(rightax_df.columns):
        set_spines(ax_right[j], which="right", loc=1+j*1.2/plt.gcf().get_size_inches()[0], color=colors[i+j+1])
        ax_right[j].set_ylabel(col,fontweight='bold',fontsize=ylabel_fontsize,loc="bottom")
        ax_right[j].plot(rightax_df.index,rightax_df[col],
                         rightax_style[j],color=colors[i+j+1],
                         linewidth=linewidth,markersize=markersize,
                         drawstyle="steps-post")
        ax_right[j].set_ylim(rightax_limits[j][0],rightax_limits[j][1])
    return ax_left, ax_right


class first_order_model:
    '''
        First order self-regualting model with dead-time
        Kp:         process gain
        Tc:         time constant
        delay:      dead-time
        dt:         sampling rate
        y0:         process output bias
        u0:         process input bias
    '''

    def __init__(self, Kp, Tc, delay, dt, y0, u0):
        self.Kp     = Kp
        self.Tc     = Tc
        self.delay  = delay
        self.dt     = dt
        self.y      = 0
        self.dy_dt  = 0
        self.y0     = y0
        self.dy_dt  = 0
        self.u0     = u0
        self.u_hist = [u0 for _ in range(int(round(delay/dt))+1)]

    def rk4(self, u):
        def dy_dt(y):
            return u*self.Kp/self.Tc - y/self.Tc

        fa = dy_dt(self.y,              )
        fb = dy_dt(self.y + fa*self.dt/2)
        fc = dy_dt(self.y + fb*self.dt/2)
        fd = dy_dt(self.y + fc*self.dt  )

        self.y = self.y + self.dt/6*(fa + 2*fb + 2*fc + fd)

    def gen_response(self, u):

        # update model for Kp and Tc updates
        self.u0 = self.u_hist[0] - (self.dy_dt + self.y/self.Tc)/(self.Kp/self.Tc)

        # update model for delay updates
        new_shape = int(round(self.delay/self.dt))+1
        old_shape = len(self.u_hist)

        if new_shape > old_shape: 
            self.u_hist.extend([self.u_hist[-1] for _ in range(new_shape - old_shape)])
        else:
            self.u_hist = self.u_hist[:new_shape]

        # update history with new u
        for i in range(len(self.u_hist)-1):
            self.u_hist[i] = self.u_hist[i+1]
        self.u_hist[-1] = u

        # generate new response
        self.rk4(self.u_hist[0] - self.u0)

        # reset dy_dt to update u0 accordingly on next execute
        self.dy_dt = (self.u_hist[0] - self.u0)*self.Kp/self.Tc - self.y/self.Tc

        return self.y + self.y0

class integrator_model:
    '''
        First order integrator model with dead-time
        K:          process gain
        delay:      dead-time
        dt:         sampling rate
        y0:         process output bias
        u:          process input (outflow-inflow)
    '''

    def __init__(self, K, delay, dt, y0, u0):
        self.K      = K
        self.delay  = delay
        self.dt     = dt
        self.y      = 0
        self.dy_dt  = 0
        self.y0     = y0
        self.dy_dt  = 0
        self.u_hist = [u0 for _ in range(int(round(delay/dt))+1)]
        self.u0     = u0

    def rk4(self, u):
        def dy_dt(y):
            return (u-self.u0)*self.K

        fa = dy_dt(self.y,              )
        fb = dy_dt(self.y + fa*self.dt/2)
        fc = dy_dt(self.y + fb*self.dt/2)
        fd = dy_dt(self.y + fc*self.dt  )
        
        self.y = self.y + self.dt/6*(fa + 2*fb + 2*fc + fd)

    def gen_response(self, u):
        # update model for delay updates
        new_shape = int(round(self.delay/self.dt))+1
        old_shape = len(self.u_hist)

        if new_shape > old_shape: 
            self.u_hist.extend([self.u_hist[-1] for _ in range(new_shape - old_shape)])
        else:
            self.u_hist = self.u_hist[:new_shape]
        
        # update history with new u
        for i in range(len(self.u_hist)-1):
            self.u_hist[i] = self.u_hist[i+1]
        self.u_hist[-1] = u
        
        # generate new response
        self.rk4(self.u_hist[0])
        
        return self.y + self.y0


from .blocks import io
from .controllers import fuzzy, pidff, gain_update
from .datagen import random_walk
from .models import fopdt, ramp
from .plotting import multi_yax_plot, update_legend, view_fuzzy_design
from .signalprocessing import kalmanfilter, lag_filter
from .systemid import dataIsolate, auto_id, TransferFunctionFit


def update_legend(ax, colors=None, x_loc=1.05, y_loc=1, linewidth=2,frameon=False):
    leg = ax.legend(loc = "upper left", bbox_to_anchor=(x_loc, y_loc), 
                    fontsize=10, frameon = frameon, edgecolor = "k", facecolor = "white")

    if colors is not None:
        for i,(c,l,text) in enumerate(zip(colors,leg.legendHandles,leg.get_texts())):
            text.set_color(c)
            l.set_linewidth(linewidth)
    else:
        for l,text in zip(leg.legendHandles,leg.get_texts()):
            try: text.set_color(l.get_color())
            except: continue
            text.set_fontweight('bold')


class pidff:

    def __init__(self, cv, sp, mv, mv_min, mv_max, dt, kc, ti, td=0, db=0, dv={}, kf={}):

        '''
         cv      controlled variable
         sp      set point
         mv      manipulated variable
         mv_min  manipulated variable lower limit
         mv_max  manipulated variable upper limit
         dt      sampling rate
         kc      controller gain
         ti      controller integral time constant
         td      controller derivative constant
         db      dead band
         dv      disturbance variable
         kf      feed forward gain (proportional)
        '''
        self.e      = [sp-cv]*3
        self.mv     = mv 
        self.mv_min = mv_min
        self.mv_max = mv_max
        self.dt     = dt
        self.kc     = kc
        self.ti     = ti
        self.td     = td 
        self.db     = db 
        self.dv     = dict(dv)
        self.kf     = dict(kf)

    def update_tune(self, mv_min, mv_max, kc, ti, td=0, db=0, kf={}):
        self.mv_min = mv_min
        self.mv_max = mv_max
        self.kc     = kc
        self.ti     = ti
        self.td     = td 
        self.db     = db 
        self.kf     = kf
        
    def update_mv(self, cv, sp, dv={}):
        self.e=[sp-cv,self.e[0],self.e[1]]
        if (abs(self.e[0]) > self.db*0.5) or self.db==0:                           # take control action if error falls outside deadband
            self.mv += (self.kc*                (self.e[0]-  self.e[1]          )+ #      P
                        self.kc/self.ti*self.dt*(self.e[0]                      )+ #      I
                        self.kc*self.td/self.dt*(self.e[0]-2*self.e[1]+self.e[2])) #      D
        for key in dv.keys():
            self.mv += self.kf[key]*(dv[key]-self.dv[key])                         #      FF
            self.dv[key]=dv[key]

        self.mv = max(min(self.mv,self.mv_max),self.mv_min)                        # clamp mv
        
        return self.mv

if __name__ == "__main__":
    import os
    import sys
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from models.ramp import integrator_model
    import numpy as np
    import matplotlib.pyplot as plt

    N  =  400
    dt =  2
    k  = -0.01
    kd =  0.3
    de =  5

    y0 = 10
    cv = y0
    sp = y0+5
    mv = 50
    dv = {"1":10,
          "2":10}

    mv_min = -10000
    mv_max = 100000
    kc     = -10
    ti     = 20
    td     = 0 
    db     = 0
    kf     = {"1":-kd/k*1,
              "2":-kd/k*1}

    m =      integrator_model(k, de,dt,0,mv)
    d = {"1":integrator_model(kd,de,dt,0,dv["1"]),
         "2":integrator_model(kd,de,dt,0,dv["2"])}

    c = pidff(cv, sp, mv, mv_min, mv_max, dt, kc, ti, td, db, dv, kf)

    T=np.arange(0,N*dt,dt)
    CV=[];SP=[];MV=[];DV={"1":[],"2":[]}
    np.random.seed(1)
    for t in T:
        dv["1"] += np.random.normal(0,0.1)
        dv["2"] += np.random.normal(0,0.2)
        sp      += np.random.normal(0,0.0)

        c.update_tune(mv_min, mv_max, kc, ti, td, db, kf)
        mv  = c.update_mv(cv, sp, dv)
        cv  = m     .gen_response(mv     )+\
              d["1"].gen_response(dv["1"])+\
              d["2"].gen_response(dv["2"])+\
              y0
        CV.append(cv);SP.append(sp);MV.append(mv);DV["1"].append(dv["1"]);DV["2"].append(dv["2"])

    fig,ax=plt.subplots(3,1,figsize=(9,6),sharex=True)
    ax[0].plot(T,CV,"k-",drawstyle="steps-post",label="cv")
    ax[0].plot(T,SP,"r-",drawstyle="steps-post",label="sp")
    ax[0].plot(T,[sp-db*.5 for sp in SP],"r-",drawstyle="steps-post",label=None)
    ax[0].plot(T,[sp+db*.5 for sp in SP],"r-",drawstyle="steps-post",label=None)
    ax[1].plot(T,MV,"k-",drawstyle="steps-post",label="mv")
    ax[2].plot(T,DV["1"],"k-",drawstyle="steps-post",label="dv1")
    ax[2].plot(T,DV["2"],"k:",drawstyle="steps-post",label="dv2")
    ax[0].legend(loc="upper right",bbox_to_anchor=(1.1,1))
    ax[1].legend(loc="upper right",bbox_to_anchor=(1.1,1))    
    ax[2].legend(loc="upper right",bbox_to_anchor=(1.1,1))
    plt.show()


import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from systemid.dataIsolate import dataIsolate
from systemid.TransferFunctionFit import TransferFunctionFit
from plotting.multi_yax_plot import multi_yax_plot
from plotting.update_legend import update_legend
import datetime
import matplotlib.pyplot as plt
import seaborn as sns

def auto_id(data, 
             aggPeriod=30, 
             optimize_iters = 1, 
             inpt="", 
             outp="", 
             highSTDdict={}, 
             lowSTDdict={},
             num=[1], 
             den=[20,0], 
             delay=0,
             minute_shift=datetime.timedelta(minutes=0),
             period_index=range(2),
             y0 = None):

    #--------------------------------------------------------------------------
    s = dataIsolate(data        = data,
                    aggPeriod   = aggPeriod,
                    highSTDdict = highSTDdict,
                    lowSTDdict  = lowSTDdict,
                    ).round(4)  
    
    print (s.shape)
    if period_index is None: period_index = range(5)
    s = s.iloc[period_index,:]
    print (s)
    fig, ax = plt.subplots(s.shape[0],2, sharey=False, figsize=(25,12))
    plt.subplots_adjust(wspace=0.55)
    
    for i,idx in enumerate(s.index):
        mask = (data.index > (s.loc[idx,"startDateTime"]) + minute_shift) &\
               (data.index < (s.loc[idx,"endDateTime"  ]) + minute_shift)
        di = data.loc[mask,list(highSTDdict.keys())+list(lowSTDdict.keys())].fillna(method="pad")
        di = di.loc[~di.isnull().any(axis=1),:]
        
        #----------------------------------------------------------------------
        tff = TransferFunctionFit(num,den,delay)
        tff.optimize(di[inpt],di[outp],optimize_iters)
        y0_ = y0
        if y0 is None: y0_ = di[outp][0]
        y = tff.gen_response(di[inpt], y0_)
    
        ax[i,0].plot(di.index, di[inpt],"k-", label = inpt,    drawstyle = "steps-post", linewidth  = 1.0)
        if i==0:update_legend(ax[i,0], ["k"],0,1.2,2,True)
        ax[i,0] = ax[i,0].twinx()
        ax[i,0].plot(di.index, di[outp],"r.", label = outp,    drawstyle = "steps-post", markersize = 4.0)
        ax[i,0].plot(di.index, y,       "r-", label = "model", drawstyle = "steps-post", linewidth  = 2.0)
        ax[i,0].set_ylim(min(min(y),di[outp].min()),max(max(y),di[outp].max()))
        if i==0:update_legend(ax[i,0],["r","r"],0.8,1.2,2,True)
        #----------------------------------------------------------------------
        text = F"num: { [round(x,6) for x in tff.num]}\n"  +\
               F"den:  {[int(round(x,0)) for x in tff.den]}\n"  +\
               F"del:   {tff.delay}\n"+\
               F"R2:    {round(tff.R2,2)}"
        ax[i,0].text(di.index[0]+(di.index[-1]-di.index[0])*1.15, 
                     di[outp].max(), 
                     text,ha='left',va='top', 
                     fontweight="bold",fontsize=12,color="white", 
                     bbox=dict(facecolor='firebrick',alpha=0.8,edgecolor='white'))
        #----------------------------------------------------------------------
        lims = []
        for lowSTD_tag in lowSTDdict.keys():
            min_   = data[lowSTD_tag].min()
            max_   = data[lowSTD_tag].max()
            range_ = max_ - min_
            lims.append([min_- 0.05*range_, max_ + 0.05*range_])
        leftax_df      = di[list(lowSTDdict.keys())].iloc[:,:3]
        rightax_df     = di[list(lowSTDdict.keys())].iloc[:,3:]
        leftax_limits  = lims[:3]
        rightax_limits = lims[3:]
        
        multi_yax_plot(ax[i,1],
                       leftax_df=leftax_df,
                       leftax_limits=leftax_limits,
                       rightax_df=rightax_df,
                       rightax_limits=rightax_limits,
                       colors=sns.color_palette("deep", len(lowSTDdict.keys())),linewidth=1)
    plt.show()

if __name__ == "__main__":
    import pandas as pd
    from datagen.random_walk import gen_data

    rows = 3601
    data = pd.DataFrame(
        gen_data(tag_info_dict={"y":{"start":0,"std":10},"u":{"start":0,"std":1},"d":{"start":0,"std":1}}, rows=rows),
        index=pd.date_range(start="2022-06-01",end="2022-06-02 01:00:00", periods = rows).to_pydatetime().tolist())

    data.loc[data.index > "2022-06-01 15:20:00", "u"] = data.loc[data.index > "2022-06-01 15:20:00", "u"] + 100
    data.plot()

    auto_id(data, 
             aggPeriod=120, 
             optimize_iters = 1, 
             inpt="u", 
             outp="y", 
             highSTDdict={"y":0,"u":0},
             lowSTDdict={"d":float("inf")},
             num=[1], 
             den=[20,1], 
             delay=0,
             minute_shift=datetime.timedelta(minutes=0),
             period_index=range(2),
             y0 = None)


def gain_update(kc_small,kc_large,cv,cv_limits=[None,None,None,None]):
    '''

    ------------------------------------------------
     positive kc
    ------------------------------------------------
    ...lim[0]                 lim[3]...  -> kc_large
             .               .
              .             .
              lim[1]...lim[2]            -> kc_small


    ------------------------------------------------
      negative kc
    ------------------------------------------------
              lim[1]...lim[2]            -> kc_small
              .             .
             .               .
    ...lim[0]                 lim[3]...  -> kc_large

    '''
    if   cv< cv_limits[0] or  cv>cv_limits[3]:return kc_large
    elif cv> cv_limits[1] and cv<cv_limits[2]:return kc_small
    elif cv<=cv_limits[1]:return (kc_small-kc_large)*(cv-cv_limits[1])/(cv_limits[1]-cv_limits[0])+kc_small
    else:                 return (kc_large-kc_small)*(cv-cv_limits[3])/(cv_limits[3]-cv_limits[2])+kc_large

if __name__ == "__main__":

    import numpy as np
    import matplotlib.pyplot as plt

    levels     = np.arange(0,100,0.1)
    level_low  = 20
    level_high = 80
    kc_small   = -0.1
    kc_large   = -1
    gains      = []
    
    for level in levels:
        gains.append(gain_update(kc_small  = kc_small, 
                                 kc_large  = kc_large, 
                                 cv        = level, 
                                 cv_limits =[level_low  - (level_high-level_low)*0.1,
                                             level_low  + (level_high-level_low)*0.1,
                                             level_high - (level_high-level_low)*0.1,
                                             level_high + (level_high-level_low)*0.1]))
    fig, ax = plt.subplots(1,1,figsize=(7,7))
    ax.plot(levels,gains,"k")
    ax.plot([level_low, level_low ],[kc_small,kc_large],"r:")
    ax.plot([level_high,level_high],[kc_small,kc_large],"r:") 
    plt.show()       


import datetime

import numpy as np
from scipy import signal as si
from scipy.optimize import curve_fit

import matplotlib.pyplot as plt
import seaborn as sns

from plotting import multi_yax_plot,update_legend

def dataIsolate(data, aggPeriod, 
                highSTDdict,
                lowSTDdict):
    '''
        function to isolate data peiods for SISO System Identification
        
        Parameters
        ----------
        data : pandas - DataFrame
            Input DataFrame
        
        aggPeriod : int - minutes
            Period over which to aggregate data for std - minutes
            
        highSTDdict : dictionary
            keys:   tags that need to have a high STD
            values: min STD limit allowed for each tag
    
        lowSTDdict : dictionary
            keys:   tags that need to have a low STD 
            values: max STD limit allowed for each tag
    
        Returns
        -------
        s : pandas - DataFrame
        columns - start & end DateTimes sorted decending according to 
        highSTDtags calculated STD's
    '''
    
    tags = list(highSTDdict.keys()) + list(lowSTDdict.keys())
    s = data[tags].resample(str(aggPeriod)+"T", closed="left").std()
    
    s["mask"] = True
    for key in highSTDdict.keys():s["mask"] = s["mask"] & (s[key]>highSTDdict[key])
    for key in lowSTDdict.keys(): s["mask"] = s["mask"] & (s[key]<lowSTDdict [key])
           
    s = s.loc[s["mask"],:]
    s = s.sort_values(by = [list(highSTDdict.keys())[0]], ascending = 0).drop(["mask"],axis=1)
    
    s["startDateTime"] = s.index
    s["endDateTime"]   = s.index + datetime.timedelta(minutes=aggPeriod)
    s = s.reset_index().iloc[:,1:]

    return s

class TransferFunctionFit:
    def __init__(self, num, den, delay):
        self.num   = num   # transfer function numerator:   list
        self.den   = den   # transfer function denominator: list
        self.delay = delay # input-output delay:            int (seconds)
        self.R2    = None
        
    def gen_response(self, u, y0):
        # sampling interval in seconds
        dt = np.median((u.index[1:]-u.index[:-1]).seconds)
        
        # Delay input
        u_shift = np.repeat(u[0], int(round(self.delay/dt))+1)
        u       = np.array((u_shift.tolist() + u.values.tolist())\
                           [:u.shape[0]])-u[0]
            
        # Discrete time intervals            
        t  = np.arange(0,u.shape[0]*dt, dt)
        
        # Generate output   
        return si.TransferFunction(self.num,self.den).output(u,t)[1]+y0
    
    def optimize(self, u, y, iterations=1):
        for _ in range(iterations):
            
            # Set intial parameters and optimization bounds
            #------------------------------------------------------------------
            params_0 = self.num + self.den
            l_bounds = []
            u_bounds = []
            for el in params_0:
                if   el==0:l_bounds.append(-1e-100); u_bounds.append(1e-100 )  
                elif el>0 :l_bounds.append(el*0.01); u_bounds.append(el*1000)
                else      :l_bounds.append(el*1000.);u_bounds.append(el*.01 )
            if round(params_0[-1],1)==1: 
                l_bounds[-1]=1
                u_bounds[-1]=1+1e-15 
            if round(params_0[-2],2)==1 and round(params_0[-1],1)==0:
                l_bounds[-2]=1
                u_bounds[-2]=1+1e-15 
            
            # Fit numerator and denominator
            #------------------------------------------------------------------
            def model(u, *params):
                self.num   = params[:len(self.num)]
                self.den   = params[len(self.num):]
                return self.gen_response(u, y[0])
            params = curve_fit(model, u, y, 
                               p0 = params_0, 
                               bounds=(l_bounds,u_bounds))[0]
            self.num = params[:len(self.num)].tolist()
            self.den = params[len(self.num):].tolist()

            # Calc Delay
            #------------------------------------------------------------------
            if self.delay != 0:
                dt = np.median((u.index[1:]-u.index[:-1]).seconds)
                c_m = {}
                for de in range(0,int(u.shape[0]*0.2*dt),int(dt)):
                    self.delay = de
                    y_m        = self.gen_response(u, y[0])
                    c_m[de]    = np.corrcoef(y_m,y)[0,1]
                self.delay = max(c_m, key=c_m.get)
            
        #----------------------------------------------------------------------
        SSres   =  np.sum((y - self.gen_response(u, y[0]))**2)
        SStot   =  np.sum((y - np.mean(y))**2)
        self.R2 = (1 - SSres/SStot)

def auto_id(data, 
             aggPeriod=30, 
             optimize_iters = 1, 
             inpt="", 
             outp="", 
             highSTDdict={}, 
             lowSTDdict={},
             num=[1], 
             den=[20,0], 
             delay=0,
             minute_shift=datetime.timedelta(minutes=0),
             period_index=range(2),
             y0 = None):

    #--------------------------------------------------------------------------
    s = dataIsolate(data        = data,
                    aggPeriod   = aggPeriod,
                    highSTDdict = highSTDdict,
                    lowSTDdict  = lowSTDdict,
                    ).round(4)  
    
    print (s.shape)
    if period_index is None: period_index = range(5)
    s = s.iloc[period_index,:]
    print (s)
    fig, ax = plt.subplots(s.shape[0],2, sharey=False, figsize=(25,12))
    plt.subplots_adjust(wspace=0.55)
    
    for i,idx in enumerate(s.index):
        mask = (data.index > (s.loc[idx,"startDateTime"]) + minute_shift) &\
               (data.index < (s.loc[idx,"endDateTime"  ]) + minute_shift)
        di = data.loc[mask,list(highSTDdict.keys())+list(lowSTDdict.keys())].fillna(method="pad")
        di = di.loc[~di.isnull().any(axis=1),:]
        
        #----------------------------------------------------------------------
        tff = TransferFunctionFit(num,den,delay)
        tff.optimize(di[inpt],di[outp],optimize_iters)
        y0_ = y0
        if y0 is None: y0_ = di[outp][0]
        y = tff.gen_response(di[inpt], y0_)
    
        ax[i,0].plot(di.index, di[inpt],"k-", label = inpt,    drawstyle = "steps-post", linewidth  = 1.0)
        if i==0:update_legend(ax[i,0], ["k"],0,1.2,2,True)
        ax[i,0] = ax[i,0].twinx()
        ax[i,0].plot(di.index, di[outp],"r.", label = outp,    drawstyle = "steps-post", markersize = 4.0)
        ax[i,0].plot(di.index, y,       "r-", label = "model", drawstyle = "steps-post", linewidth  = 2.0)
        ax[i,0].set_ylim(min(min(y),di[outp].min()),max(max(y),di[outp].max()))
        if i==0:update_legend(ax[i,0],["r","r"],0.8,1.2,2,True)
        #----------------------------------------------------------------------
        text = F"num: { [round(x,6) for x in tff.num]}\n"  +\
               F"den:  {[int(round(x,0)) for x in tff.den]}\n"  +\
               F"del:   {tff.delay}\n"+\
               F"R2:    {round(tff.R2,2)}"
        ax[i,0].text(di.index[0]+(di.index[-1]-di.index[0])*1.15, 
                     di[outp].max(), 
                     text,ha='left',va='top', 
                     fontweight="bold",fontsize=12,color="white", 
                     bbox=dict(facecolor='firebrick',alpha=0.8,edgecolor='white'))
        #----------------------------------------------------------------------
        lims = []
        for lowSTD_tag in lowSTDdict.keys():
            min_   = data[lowSTD_tag].min()
            max_   = data[lowSTD_tag].max()
            range_ = max_ - min_
            lims.append([min_- 0.05*range_, max_ + 0.05*range_])
        leftax_df      = di[list(lowSTDdict.keys())].iloc[:,:3]
        rightax_df     = di[list(lowSTDdict.keys())].iloc[:,3:]
        leftax_limits  = lims[:3]
        rightax_limits = lims[3:]
        
        multi_yax_plot(ax[i,1],
                       leftax_df=leftax_df,
                       leftax_limits=leftax_limits,
                       rightax_df=rightax_df,
                       rightax_limits=rightax_limits,
                       colors=sns.color_palette("deep", len(lowSTDdict.keys())),linewidth=1)
    plt.show()


class lag_filter:

    def __init__(self, signal, tc, dt):
        '''
        Calculates the 1st order lag of an input signal
            signal:  current signal
            fsignal: previous filtered signal
            tc:      lag/time constant
            dt:      sampling rate
        '''
        self.fsignal = signal
        self.tc = tc
        self.dt = dt
        
    def update(self, signal, tc=None, dt=None):
        if tc is not None: self.tc = tc
        if dt is not None: self.dt = dt
        if self.dt == 0: return signal
        lamb = 1./(self.tc/self.dt+1)
        self.fsignal = (1.-lamb)*self.fsignal + lamb*signal
        return self.fsignal

if __name__ == "__main__":

    import numpy as np
    import matplotlib.pyplot as plt
    
    N = 1000
    dt= 1    
    tc= 10
    t = np.arange(0,N*dt,dt)

    signal = 2
    lf = lag_filter(signal, tc, dt)
    T=[];M=[];F=[]
    np.random.seed(1)
    for ti in t:
        if ti == N*dt*0.25: tc*=2; print (ti, tc)
        if ti == N*dt*0.50: tc*=2; print (ti, tc)
        if ti == N*dt*0.75: tc*=2; print (ti, tc)
        
        signal += np.random.normal(0,10)        # true
        msignal = signal+np.random.normal(0,80) # measured
        fsignal = lf.update(msignal,tc,dt)      # filtered
        M.append(msignal)
        T.append( signal)
        F.append(fsignal)
        
    plt.figure(figsize=(20,7))
    plt.plot(t,M,"k-",linewidth=0.5,drawstyle="steps-post",label="measured")
    plt.plot(t,T,"b-",linewidth=1.0,drawstyle="steps-post",label="true")
    plt.plot(t,F,"r-",linewidth=2.0,drawstyle="steps-post",label="filtered")
    plt.legend(loc="upper right")
    plt.show()


from flask import Flask, request, jsonify

tags = {}

app = Flask(__name__)

@app.route("/opc/update_tags", methods=["POST"])

def update_tags():
    response = request.json
    for tag in response.keys():
        tags[tag]={
            "value":response[tag]["value"],
            "timestamp":response[tag]["timestamp"],
            "quality":response[tag]["quality"]}
    return jsonify(tags)

@app.route("/opc/get_tags", methods=["GET"])
def get_tags():
    return jsonify(tags)

if __name__ == '__main__':
    app.run(debug=True)


#%%
import requests
import json
import datetime

#%%
timestamp = datetime.datetime.utcnow().isoformat()
tags = {
    "u1":{"value":1,"timestamp":timestamp,"quality":"good"},
    "u2":{"value":2,"timestamp":timestamp,"quality":"good"},
    "y1":{"value":3,"timestamp":timestamp,"quality":"good"},
    "y2":{"value":4,"timestamp":timestamp,"quality":"good"}
}

requests.post('http://127.0.0.1:5000/opc/update_tags', json=tags)
response = requests.get('http://127.0.0.1:5000/opc/get_tags')
print (json.dumps(response.json(),indent=4))

tags = {
    "y1":{"value":30,"timestamp":timestamp,"quality":"bad"},
}
requests.post('http://127.0.0.1:5000/opc/update_tags', json=tags)
response = requests.get('http://127.0.0.1:5000/opc/get_tags')
print (json.dumps(response.json(),indent=4))

# %%



# https://packaging.python.org/en/latest/tutorials/packaging-projects/

from setuptools import setup, find_packages

setup(
    name='apclib',
    version='0.0.3',
    license='MIT',
    author="Marnus Olivier",
    author_email='mo@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/MarnusOlivier/apcsim',
    keywords='apcsim',
    install_requires=["matplotlib","numpy","pandas","scipy","seaborn",],
)


import datetime

def dataIsolate(data, aggPeriod, 
                highSTDdict,
                lowSTDdict):
    '''
        function to isolate data peiods for SISO System Identification
        
        Parameters
        ----------
        data : pandas - DataFrame
            Input DataFrame
        
        aggPeriod : int - minutes
            Period over which to aggregate data for std - minutes
            
        highSTDdict : dictionary
            keys:   tags that need to have a high STD
            values: min STD limit allowed for each tag
    
        lowSTDdict : dictionary
            keys:   tags that need to have a low STD 
            values: max STD limit allowed for each tag
    
        Returns
        -------
        s : pandas - DataFrame
        columns - start & end DateTimes sorted decending according to 
        highSTDtags calculated STD's
    '''
    
    tags = list(highSTDdict.keys()) + list(lowSTDdict.keys())
    s = data[tags].resample(str(aggPeriod)+"T", closed="left").std()
    
    s["mask"] = True
    for key in highSTDdict.keys():s["mask"] = s["mask"] & (s[key]>highSTDdict[key])
    for key in lowSTDdict.keys(): s["mask"] = s["mask"] & (s[key]<lowSTDdict [key])
           
    s = s.loc[s["mask"],:]
    s = s.sort_values(by = [list(highSTDdict.keys())[0]], ascending = 0).drop(["mask"],axis=1)
    
    s["startDateTime"] = s.index
    s["endDateTime"]   = s.index + datetime.timedelta(minutes=aggPeriod)
    s = s.reset_index().iloc[:,1:]

    return s

if __name__ == "__main__":
    import pandas as pd
    import matplotlib.pyplot as plt
    import os
    import sys
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from datagen.random_walk import gen_data

    rows = 17281
    data = pd.DataFrame(
        gen_data(rows=rows),
        index=pd.date_range(start="2022-06-01",end="2022-06-02", periods = rows).to_pydatetime().tolist())
    data[["x1","x4"]].plot()
    s = dataIsolate(
        data, 
        aggPeriod=120, 
        highSTDdict={"x4":30},
        lowSTDdict={"x1":12})
    ax = plt.gca()

    for index in s.index:
        mask = (data.index > s.loc[index,"startDateTime"]) & (data.index < s.loc[index,"endDateTime"])
        ax.fill_between(x = data.index, y1=data.min().min(), y2=data.max().max(), where = mask, color = "k", alpha=0.1)
    print (s)
    plt.show()


# %%
import os
import sys
sys.path.append(
        os.path.dirname(
            os.path.dirname(__file__))+"\\src")

print (sys.path)
import apclib
from apclib import controllers
import apclib.models

# print (apclib.controllers.gain_update)
# %%
from apclib.controllers import fuzzy


# Matrix operation functions
if 1:
    def identity(n, v):
        m=[[0 for x in range(n)] for y in range(n)]
        for i in range(0,n):
            m[i][i] = v
        return m
    
    def matmult(a, b):
        if    len(a) == 1 and len(a[0]) == 1:
            zip_b = zip(*b)
            zip_b = list(zip_b)
            return [[el*a[0][0] for el in row] for row in b]
    
        elif len(b) == 1 and len(b[0]) == 1: 
            zip_a = zip(*a)
            zip_a = list(zip_a)
            return [[el*b[0][0] for el in row] for row in a]
    
        else:
            zip_b = zip(*b)
            zip_b = list(zip_b)
            return [[sum(ele_a*ele_b for ele_a, ele_b in zip(row_a, col_b))
                     for col_b in zip_b] for row_a in a]
    
    def matadd(a, b):
        res = []
        for i in range(len(a)):
            row = []
            for j in range(len(a[0])):
                row.append(a[i][j]+b[i][j])
            res.append(row)
        return res
    
    def matsubtract(a, b):
        res = []
        for i in range(len(a)):
            row = []
            for j in range(len(a[0])):
                row.append(a[i][j]-b[i][j])
            res.append(row)
        return res
    
    def mattranspose(a):
        return list(map(list, zip(*a)))
    
    def matminor(m,i,j):
        return [row[:j] + row[j+1:] for row in (m[:i]+m[i+1:])]
    
    def matdeternminant(m):  
        #base case for 2x2 matrix
        if len(m) == 2:
            return m[0][0]*m[1][1]-m[0][1]*m[1][0]
    
        determinant = 0
        for c in range(len(m)):
            determinant += ((-1)**c)*m[0][c]*matdeternminant(matminor(m,0,c))
        return determinant
    
    def matinverse(m):
        #base case for 1x1 matrix
        if len(m) == 1:
            return [[1./m[0][0]]]
            
        determinant = matdeternminant(m)
        #special case for 2x2 matrix:
        if len(m) == 2:
            return [[m[1][1]/determinant, -1*m[0][1]/determinant],
                    [-1*m[1][0]/determinant, m[0][0]/determinant]]
    
        #find matrix of cofactors
        cofactors = []
        for r in range(len(m)):
            cofactorRow = []
            for c in range(len(m)):
                minor = matminor(m,r,c)
                cofactorRow.append(((-1)**(r+c)) * matdeternminant(minor))
            cofactors.append(cofactorRow)
        cofactors = mattranspose(cofactors)
        for r in range(len(cofactors)):
            for c in range(len(cofactors)):
                cofactors[r][c] = cofactors[r][c]/determinant
        return cofactors

class KF:

    def __init__(self, X, U, A, B, H, q, r):
        '''
        X = AX + BU
        Y = HX
        '''
        self.q = q                               # Expected variance/error in model
        self.r = r                               # Expected variance/error in measurement
        
        # Matrices
        self.X = X                               # Mean state estimate of the previous step (k -1)
        self.U = U                               # Model Input
        self.A = A                               # State transition nxn matrix: Relates state in previous time-step to state in current time-step
        self.B = B                               # Input effect matrix:         Relates the input to the states
        self.H = H                               # Relates the measurement y to the states
        self.Q = identity(len(self.A[0]),self.q) # Process noise covariance matrix:     Expected variance/error in model
        self.R = identity(len(self.H   ),self.r) # Measurement noise covariance matrix: Expected variance/error in measurement
        self.P = identity(len(self.A[0]), 1)     # State covariance of previous step (k -1)
        
    def predict(self,U): 
        self.U = U
        self.X = matadd(matmult(self.A,self.X), matmult(self.B,self.U))
        self.P = matadd(matmult(self.A, matmult(self.P, mattranspose(self.A))),self.Q)

    def update(self, y):
        Y      = [[y]]
        IM     = matmult    (self.H, self.X)
        IS     = matadd     (self.R, matmult(self.H, matmult(self.P, mattranspose(self.H))))
        K      = matmult    (self.P, matmult(mattranspose(self.H), matinverse (IS)))
        self.X = matadd     (self.X, matmult(K, matsubtract(Y, IM)))
        self.P = matsubtract(self.P, matmult(K, matmult (IS, mattranspose(K))))


class tag:
    
    def __init__(self):
        self.value     = None
        self.timestamp = None
        self.quality   = None
        
    def set(self, value=None, timestamp=None, quality=None):
        self.value     = value
        self.timestamp = timestamp
        self.quality   = quality
        
class field:
    def __init__(self):
        self.tags = []

    def add_tag(self, tag_id):
        self.__setattr__(tag_id, tag())
        
    def add_tags(self, tags):
        for t in tags:
            self.add_tag(t)
            self.tags.append(t)

first_run = True
def first_execute():
    global first_run
    if first_run:
        first_run = False
        return True
    else:
        return False

def initialize_block(input_tag_list, output_tag_list):
    inputs, outputs = field(), field()
    inputs.add_tags(input_tag_list)
    outputs.add_tags(output_tag_list)
    return inputs, outputs

if __name__ == "__main__":
    input_tag_list  = ["tag1","tag2","tag3","tag4"]
    output_tag_list = ["tag5","tag6","tag7","tag8"]

    inputs, outputs = initialize_block(input_tag_list, output_tag_list)
    for t in inputs.tags:
        print (inputs.tag1.value)
    
    # print (first_execute())
    # print (first_execute())

    


class fuzzy:

    def __init__(self):
        self.input  = {}
        self.output = {}
        self.rules  = {}   

    def gen_xrange(self,io, N=100):
        '''
            generates a list of numbers between the min and max limits of all 
            the membership functions for an input|output (io)
            
            io: input[tag] | output[tag]
            N: number of increments
            return: list of number between min and max limits 
        '''
        def f_range(start, stop, inc):
            result = [start]
            while result[-1] < stop:
                result.append(result[-1]+inc)
            return result
        x = []
        for func in io["function"]:
            x.extend(io["function"][func]["limits"])
        x = f_range(min(x),max(x),(max(x)-min(x))/N)
        return x

    def memfunc(self,inpt=1,func="",limits=[0,1,2],memMax=1.):
        ''' inpt:   input
            func:   ltrap - left trapezium
                    rtrap - right trapezium
                    else  - triangle
            limits: func definition limits
            memMax: max allowed degree of membership
            
            return: degree of membership
        '''
        if   func == "ltrap" : return max(min((inpt-limits[2])/(limits[1]-limits[2]),memMax),0) # left trapesium
        elif func == "rtrap" : return max(min((limits[0]-inpt)/(limits[0]-limits[1]),memMax),0) # right trapesium
        elif inpt < limits[1]: return max(min((inpt-limits[0])/(limits[1]-limits[0]),memMax),0) # triangle left side
        else:                  return max(min((limits[2]-inpt)/(limits[2]-limits[1]),memMax),0) # triangle right side

    def gen_input_membership(self):
        '''
            generates input membership for each input on each memebership function
        '''
        for inpt in self.input.keys():
            for func in self.input[inpt]["function"].keys():
                self.input[inpt]["function"][func]["membership"] = self.memfunc(inpt  =self.input[inpt]["value"],
                                                                                func  =self.input[inpt]["function"][func]["type"],
                                                                                limits=self.input[inpt]["function"][func]["limits"],
                                                                                memMax=self.input[inpt]["function"][func]["membership_max"])

    def gen_output_membership(self):
        '''
            generates output membership for each rule on respective output memebership function
        '''
        for rule in self.rules.keys():
            membership = [self.input[inpt]["function"][self.rules[rule]["input"][inpt]]["membership"] for inpt in self.rules[rule]["input"]]
            membership = sum(membership)/float(len(membership))
            for otpt in self.rules[rule]["output"]:
                self.output[otpt]["function"][self.rules[rule]["output"][otpt]]["membership"] = min(membership,
                                                                                                    self.output[otpt]["function"][self.rules[rule]["output"][otpt]]["membership_max"]) 

    def calc_centroid(self):
        '''
            calculate the centroid for each output and assigns 
            it to the "value" key entry 
        '''
        for otpt in self.output.keys():
            x = self.gen_xrange(self.output[otpt])
            y = []
            for func in self.output[otpt]["function"].keys():
                y_ = [self.memfunc(inpt  =xi,
                                   func  =self.output[otpt]["function"][func]["type"],
                                   limits=self.output[otpt]["function"][func]["limits"],
                                   memMax=self.output[otpt]["function"][func]["membership"]) for xi in x]
                y.append(y_)
            y = [max(yi) for yi in zip(*y)]        
            try: self.output[otpt]["value"] = sum(xi*yi for xi,yi in zip(x,y))/sum(y)
            except: pass

    def gen_output(self):
        '''
            for a given set values in the inputs
            the following steps are followed:
            1. Update the input degree of membership for each membership function
            2. For each rule assign the degree of membership for each output
            3. Update the value for each output by calculating the centroid
               from the degree of membership from each membership function
        '''
        self.gen_input_membership()
        self.gen_output_membership()
        self.calc_centroid()

if __name__ == "__main__":
    import os
    import sys
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from plotting.view_fuzzy_design import view_fuzzy_design

    if 1:
        # inputs
        l_c_rc = 20.
        y_c_rc = 26.
        h_c_rc = 27.
        
        l_c_rt = 0.4
        y_c_rt = 0.75
        h_c_rt = 0.8
        
        l_c_cc = 52.
        y_c_cc = 53.
        h_c_cc = 58.
        
        # outputs
        l_f_rc = 50
        h_f_rc = 120
        dr_f_rc_m = (h_f_rc-l_f_rc)/10
        
        l_v_s1 = 5
        h_v_s1 = 15
        dr_v_s1_m = (h_v_s1-l_v_s1)/10
        
        l_f_cc = 3
        h_f_cc = 10
        dr_f_cc_m = (h_f_cc-l_f_cc)/10

    fuzzy_c = fuzzy()

    # inputs
    p = [0.,.1,.3,.5,.7,.9,1.]
    fuzzy_c.input["y_c_rt"]   = {"value"   : y_c_rt, 
                                 "function":{"low"     :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[l_c_rt+(h_c_rt-l_c_rt)*p[0], l_c_rt+(h_c_rt-l_c_rt)*p[0], l_c_rt+(h_c_rt-l_c_rt)*p[2]]},
                                             "good"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[l_c_rt+(h_c_rt-l_c_rt)*p[1], l_c_rt+(h_c_rt-l_c_rt)*p[3], l_c_rt+(h_c_rt-l_c_rt)*p[5]]},
                                             "high"    :{"membership":0, "membership_max":1, "type":"rtrap", "limits":[l_c_rt+(h_c_rt-l_c_rt)*p[4], l_c_rt+(h_c_rt-l_c_rt)*p[6], l_c_rt+(h_c_rt-l_c_rt)*p[6]]}}}

    fuzzy_c.input["y_c_rc"]   = {"value"    : y_c_rc, 
                                 "function":{"low"     :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[l_c_rc+(h_c_rc-l_c_rc)*p[0], l_c_rc+(h_c_rc-l_c_rc)*p[0], l_c_rc+(h_c_rc-l_c_rc)*p[2]]},
                                             "good"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[l_c_rc+(h_c_rc-l_c_rc)*p[1], l_c_rc+(h_c_rc-l_c_rc)*p[3], l_c_rc+(h_c_rc-l_c_rc)*p[5]]},
                                             "high"    :{"membership":0, "membership_max":1, "type":"rtrap", "limits":[l_c_rc+(h_c_rc-l_c_rc)*p[4], l_c_rc+(h_c_rc-l_c_rc)*p[6], l_c_rc+(h_c_rc-l_c_rc)*p[6]]}}}

    fuzzy_c.input["y_c_cc"]   = {"value"    : y_c_cc, 
                                 "function":{"low"     :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[l_c_cc+(h_c_cc-l_c_cc)*p[0], l_c_cc+(h_c_cc-l_c_cc)*p[0], l_c_cc+(h_c_cc-l_c_cc)*p[2]]},
                                             "good"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[l_c_cc+(h_c_cc-l_c_cc)*p[1], l_c_cc+(h_c_cc-l_c_cc)*p[3], l_c_cc+(h_c_cc-l_c_cc)*p[5]]},
                                             "high"    :{"membership":0, "membership_max":1, "type":"rtrap", "limits":[l_c_cc+(h_c_cc-l_c_cc)*p[4], l_c_cc+(h_c_cc-l_c_cc)*p[6], l_c_cc+(h_c_cc-l_c_cc)*p[6]]}}}

    # outputs    
    fuzzy_c.output["dr_f_rc"] = {"value"   : 0, 
                                 "function":{"reduce"  :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[-dr_f_rc_m*2.09, -dr_f_rc_m, dr_f_rc_m     ]},
                                             "none"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[-dr_f_rc_m,       0,         dr_f_rc_m     ]},
                                             "increase":{"membership":0, "membership_max":1, "type":"rtrap", "limits":[-dr_f_rc_m,       dr_f_rc_m, dr_f_rc_m*2.09]}}}                                       

    fuzzy_c.output["dr_v_s1"] = {"value"   : 0, 
                                 "function":{"reduce"  :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[-dr_v_s1_m*2.09, -dr_v_s1_m, dr_v_s1_m     ]},
                                             "none"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[-dr_v_s1_m,       0,         dr_v_s1_m     ]},
                                             "increase":{"membership":0, "membership_max":1, "type":"rtrap", "limits":[-dr_v_s1_m,       dr_v_s1_m, dr_v_s1_m*2.09]}}}                                          
                           
    fuzzy_c.output["dr_f_cc"] = {"value"   : 0, 
                                 "function":{"reduce"  :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[-dr_f_cc_m*2.09, -dr_f_cc_m, dr_f_cc_m     ]},
                                             "none"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[-dr_f_cc_m,       0,         dr_f_cc_m     ]},
                                             "increase":{"membership":0, "membership_max":1, "type":"rtrap", "limits":[-dr_f_cc_m,       dr_f_cc_m, dr_f_cc_m*2.09]}}}                               

    # rules
    #-----------------------------------------------------------
    fuzzy_c.rules["f_rc1"] = {"input" :{"y_c_rt" :"high", 
                                        "y_c_rc" :"high"},
                              "output":{"dr_f_rc":"increase"}}

    fuzzy_c.rules["f_rc2"] = {"input" :{"y_c_rt" :"low", 
                                        "y_c_rc" :"low"},
                              "output":{"dr_f_rc":"reduce"}}

    fuzzy_c.rules["f_rc3"] = {"input" :{"y_c_rt" :"good",
                                        "y_c_rc" :"good"},
                              "output":{"dr_f_rc":"none"}}
    #-----------------------------------------------------------
    fuzzy_c.rules["v_rt1"] = {"input" :{"y_c_rt" :"high"},
                              "output":{"dr_v_s1":"increase"}}
    
    fuzzy_c.rules["v_rt2"] = {"input" :{"y_c_rt" :"low"},
                              "output":{"dr_v_s1":"reduce"}}

    fuzzy_c.rules["v_rt3"] = {"input" :{"y_c_rt" :"good"},
                              "output":{"dr_v_s1":"none"}}
    #-----------------------------------------------------------
    fuzzy_c.rules["f_cc1"] = {"input" :{"y_c_cc" :"high"},
                              "output":{"dr_f_cc":"increase"}}
    
    fuzzy_c.rules["f_cc2"] = {"input" :{"y_c_cc" :"low"},
                              "output":{"dr_f_cc":"reduce"}}

    fuzzy_c.rules["f_cc3"] = {"input" :{"y_c_cc" :"good"},
                              "output":{"dr_f_cc":"none"}}
    #-----------------------------------------------------------
    view_fuzzy_design(fuzzy_c)
    import matplotlib.pyplot as plt
    plt.show()


import numpy as np
np.random.seed(1)

class random_walk:
    '''
        random walk data generation class
    '''
    def __init__(self, start, std):
        self.value = start
        self.std   = std

    def walk(self):
        self.value += np.random.normal(0,self.std)
        return self.value

def gen_data(tag_info_dict=None, rows=1000):
    
    if tag_info_dict is None:
        tag_info_dict = {}
        for i,tag in enumerate(["x1", "x2", "x3", "x4"]):
            tag_info_dict[tag] = {"start":i+1, "std": i+1}

    data_gen_objects = {}
    data = {}
    for tag in tag_info_dict.keys():
        data_gen_objects[tag] = random_walk(tag_info_dict[tag]["start"],tag_info_dict[tag]["std"])
        data[tag] = []
    
    for i in range(rows):
        for tag in tag_info_dict.keys():
            data[tag].append(data_gen_objects[tag].value)
            data_gen_objects[tag].walk()

    return data


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def view_fuzzy_design(fuzzy):
    plt.rcParams.update({"lines.color":       "white",
                         "patch.edgecolor":   "white",
                         "text.color":        "white",
                         "axes.edgecolor":    "white",
                         "axes.labelcolor":   "white",
                         "xtick.color":       "white",
                         "ytick.color":       "white",
                         "axes.grid" :         False, 
                         "grid.color":        "lightgray",
                         "axes.facecolor":    "darkslategray",
                         "figure.facecolor":  "k",
                         "figure.edgecolor":  "k",
                         "savefig.facecolor": "k",
                         "savefig.edgecolor": "k",
                         'legend.handlelength':0})
    
    '''
        fuzzy: instance of fuzzy class
    '''
    num_rows = max(len(fuzzy.input.keys()), len(fuzzy.output.keys()))
    if num_rows == 1:num_rows = 2
    
    fig, ax = plt.subplots(num_rows, 3, figsize = (20, num_rows*3))
    plt.subplots_adjust(wspace=0.4, hspace=0.4)
    colors = ["b","g","r","c","m","y","k","w"]*100
    colors = ["c","y","tomato"]*100
    
    # Plot input & output memebership functions
    fuzzy.gen_output()    
    for j,io in enumerate([fuzzy.input, fuzzy.output]):
        
        for i,key in enumerate(io.keys()):
            x = fuzzy.gen_xrange(io[key])
            for k,func in enumerate(io[key]["function"]):
                y = [fuzzy.memfunc(inpt  =xi,
                                   func  =io[key]["function"][func]["type"],
                                   limits=io[key]["function"][func]["limits"],
                                   memMax=io[key]["function"][func]["membership_max"]) for xi in x]
                ax[i,j].plot(x,y,color=colors[k],label=func)
                ax[i,j].text(io[key]["value"],
                             io[key]["function"][func]["membership"], 
                             str(round(io[key]["function"][func]["membership"],2)),
                             color="w",fontweight="bold",
                             ha="center", va="center", bbox=dict(facecolor=colors[k],edgecolor='none',pad=1.0))
                if j > 0:
                    y = [fuzzy.memfunc(inpt  =xi,
                                       func  =io[key]["function"][func]["type"],
                                       limits=io[key]["function"][func]["limits"],
                                       memMax=io[key]["function"][func]["membership"]) for xi in x]
                    ax[i,j].fill_between(x,y,color=colors[k],alpha=0.4)
            leg = ax[i,j].legend(loc="upper center", ncol=len(io[key]["function"].keys()),handletextpad=0.,frameon=True)
            for c,l,t in zip(colors,leg.legendHandles,leg.get_texts()):t.set_color(c);t.set_ha('left');t.set_fontweight('bold');l.set_linewidth(0)
            ax[i,j].set_ylim(-0.1,1.55)
            ax[i,j].set_yticks([0,.2,.4,.6,.8,1])
            ax[i,j].plot([io[key]["value"],io[key]["value"]],[0,1],"w--",linewidth=1)
            ax[i,j].text(io[key]["value"], 1.08, str(round(io[key]["value"],2)),fontsize=11,ha="center") 
            title = "INPUT"
            if j > 0: title = "OUTPUT" 
            ax[i,j].set_title(title+": "+key)
            
    # Simulate multiple scenarios
    io_links = {}
    for rule in fuzzy.rules.keys():
        for otpt in fuzzy.rules[rule]["output"].keys():
            if otpt not in io_links.keys():io_links[otpt] = []
            for inpt in fuzzy.rules[rule]["input"].keys():
                if inpt not in io_links[otpt]: io_links[otpt].append(inpt)
    for i,otpt in enumerate(io_links.keys()):
        x = {}
        for inpt,N in zip(io_links[otpt][:2],[50,10]): x[inpt] = fuzzy.gen_xrange(fuzzy.input[inpt],N)
  
        if len(x.keys()) == 1:
            inpt = list(x.keys())[0]
            y = []
            for xi in x[inpt]:
                fuzzy.input[inpt]["value"] = xi
                fuzzy.gen_output()
                y.append(fuzzy.output[otpt]["value"])
            ax[i,2].plot(x[inpt],y,"w",drawstyle="steps-post")
            ax[i,2].set_ylabel("OUTPUT: "+otpt)
            ax[i,2].set_xlabel("INPUT: " +inpt)
        else:
            low_len_key = [(key, len(x[key])) for key in x.keys()]
            low_len_key.sort(key=lambda x: x[1])
            low_len_key, high_len_key  = low_len_key[0][0],low_len_key[1][0]
            
            colors = sns.color_palette("Spectral",len(x[low_len_key]))[::-1]
            y_prev = []
            for k,xi in enumerate(x[low_len_key]):
                fuzzy.input[low_len_key]["value"] = xi
                y = []
                for xii in x[high_len_key]:
                    fuzzy.input[high_len_key]["value"] = xii
                    fuzzy.gen_output()
                    y.append(fuzzy.output[otpt]["value"])
                if k==0 or sum([(abs(yi/np.std(y)-y_previ/np.std(y_prev))) for yi,y_previ in zip(y,y_prev)]) > 1e-10:
                    ax[i,2].plot(x[high_len_key],y,color=colors[k],label=str(round(xi,2)),drawstyle="steps-post")
                    ax[i,2].text(x[high_len_key][int(len(y)*.5)],y[int(len(y)*.5)],str(round(xi,2)),
                                 color=colors[k],fontweight="bold",
                                 ha="center", va="center", bbox=dict(facecolor="k",edgecolor='none',pad=1.0,alpha=1))
                y_prev = y
            ax[i,2].set_ylabel("OUTPUT: "+otpt)
            ax[i,2].set_xlabel("INPUT: " +high_len_key)
            leg = ax[i,2].legend(loc="upper right", 
                                 frameon = False, 
                                 title = "INPUT: " + low_len_key, 
                                 bbox_to_anchor=(1.4, 1.),
                                 ncol=int(len(ax[i,2].lines)>5)+1,
                                 handletextpad=0.)
            for l,t in zip(leg.legendHandles,leg.get_texts()):
                t.set_color(l.get_color())
                t.set_ha('center')
                t.set_fontweight('bold')
                l.set_linewidth(0)
            leg.get_title().set_position((-15, 0))
            leg._legend_box.align = "left"
    
    # Clean empty plots
    for i in range(ax.shape[0]):
        for j in range(ax.shape[1]):
            if len(ax[i,j].lines) == 0:
                fig.delaxes(ax[i,j])

def multi_yax_plot(ax, 
                   leftax_df=pd.DataFrame(),  # left  axis data frame 
                   rightax_df=pd.DataFrame(), # right axis data frame
                   leftax_limits=None,        # options include:
                   rightax_limits=None,       # None | [low,high] | [[low,high], [low,high], [low,high]]
                   leftax_style="-",          # options include:
                   rightax_style="-",         # "-", ".", ".-" | ["-",".",":"]
                   colors="bright",           # "palette" | ["k", "r", "g"]
                   linewidth=None,
                   markersize=None,
                   ylabel_fontsize=15):
    
    def set_spines(ax,which="left",loc=0,color="k"):
        if which!= "host":
            left_right = ["right","left"]
            if which == "left": left_right = ["left","right"]
            ax.spines[left_right[0]].set_position(("axes", loc))
            ax.spines[left_right[0]].set_visible(True)
            ax.spines[left_right[0]].set_linewidth(3)
            ax.spines[left_right[1]].set_visible(False)
            ax.spines["bottom"].     set_visible(False)
            ax.spines["top"].        set_visible(False)
            plt.setp(ax.spines.values(), color=color)
            ax.yaxis.set_label_position(which)
            ax.yaxis.set_ticks_position(which)
            ax.yaxis.label.set_color(color)
            ax.tick_params(axis='y', colors=color,width=2,size=5)
        else:
            ax.axes.get_yaxis().set_visible(False)
            ax.spines["right"].set_visible(False)
            ax.spines["left"].set_visible(False)
            ax.spines["top"].set_visible(False)
            ax.spines["bottom"].set_visible(True)
    
    if leftax_limits  is None:leftax_limits =[[None,None] for _ in range(leftax_df. shape[1])]
    if rightax_limits is None:rightax_limits=[[None,None] for _ in range(rightax_df.shape[1])]
    if not any(isinstance(x, list) for x in leftax_limits): leftax_limits =[[leftax_limits[0], leftax_limits[1]] for _ in range(leftax_df. shape[1])]
    if not any(isinstance(x, list) for x in rightax_limits):rightax_limits=[[rightax_limits[0],rightax_limits[1]]for _ in range(rightax_df.shape[1])]
    if not isinstance(leftax_style, list): leftax_style =[leftax_style  for _ in range(leftax_df. shape[1])]
    if not isinstance(rightax_style,list): rightax_style=[rightax_style for _ in range(rightax_df.shape[1])]
    if not isinstance(colors,list): colors = sns.color_palette(colors, leftax_df.shape[1]+rightax_df.shape[1])
    
    set_spines(ax, which="host")
    ax_left  = [ax.twinx() for _ in range(leftax_df. shape[1])]
    ax_right = [ax.twinx() for _ in range(rightax_df.shape[1])]
    
    for i,col in enumerate(leftax_df.columns):
        set_spines(ax_left[i], which="left", loc=-(i*1.2)/plt.gcf().get_size_inches()[0], color=colors[i])
        ax_left[i].set_ylabel(col,fontweight='bold',fontsize=ylabel_fontsize,loc="bottom")
        ax_left[i].plot(leftax_df.index,leftax_df[col],
                        leftax_style[i],color=colors[i],
                        linewidth=linewidth,markersize=markersize,
                        drawstyle="steps-post")
        ax_left[i].set_ylim(leftax_limits[i][0],leftax_limits[i][1])

    if leftax_df.columns.shape[0] == 0: i=0
    for j,col in enumerate(rightax_df.columns):
        set_spines(ax_right[j], which="right", loc=1+j*1.2/plt.gcf().get_size_inches()[0], color=colors[i+j+1])
        ax_right[j].set_ylabel(col,fontweight='bold',fontsize=ylabel_fontsize,loc="bottom")
        ax_right[j].plot(rightax_df.index,rightax_df[col],
                         rightax_style[j],color=colors[i+j+1],
                         linewidth=linewidth,markersize=markersize,
                         drawstyle="steps-post")
        ax_right[j].set_ylim(rightax_limits[j][0],rightax_limits[j][1])
    return ax_left, ax_right

def update_legend(ax, colors=None, x_loc=1.05, y_loc=1, linewidth=2,frameon=False):
    leg = ax.legend(loc = "upper left", bbox_to_anchor=(x_loc, y_loc), 
                    fontsize=10, frameon = frameon, edgecolor = "k", facecolor = "white")

    if colors is not None:
        for i,(c,l,text) in enumerate(zip(colors,leg.legendHandles,leg.get_texts())):
            text.set_color(c)
            l.set_linewidth(linewidth)
    else:
        for l,text in zip(leg.legendHandles,leg.get_texts()):
            try: text.set_color(l.get_color())
            except: continue
            text.set_fontweight('bold')


import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

def view_fuzzy_design(fuzzy):
    plt.rcParams.update({"lines.color":       "white",
                         "patch.edgecolor":   "white",
                         "text.color":        "white",
                         "axes.edgecolor":    "white",
                         "axes.labelcolor":   "white",
                         "xtick.color":       "white",
                         "ytick.color":       "white",
                         "axes.grid" :         False, 
                         "grid.color":        "lightgray",
                         "axes.facecolor":    "darkslategray",
                         "figure.facecolor":  "k",
                         "figure.edgecolor":  "k",
                         "savefig.facecolor": "k",
                         "savefig.edgecolor": "k",
                         'legend.handlelength':0})
    
    '''
        fuzzy: instance of fuzzy class
    '''
    num_rows = max(len(fuzzy.input.keys()), len(fuzzy.output.keys()))
    if num_rows == 1:num_rows = 2
    
    fig, ax = plt.subplots(num_rows, 3, figsize = (20, num_rows*3))
    plt.subplots_adjust(wspace=0.4, hspace=0.4)
    colors = ["b","g","r","c","m","y","k","w"]*100
    colors = ["c","y","tomato"]*100
    
    # Plot input & output memebership functions
    fuzzy.gen_output()    
    for j,io in enumerate([fuzzy.input, fuzzy.output]):
        
        for i,key in enumerate(io.keys()):
            x = fuzzy.gen_xrange(io[key])
            for k,func in enumerate(io[key]["function"]):
                y = [fuzzy.memfunc(inpt  =xi,
                                   func  =io[key]["function"][func]["type"],
                                   limits=io[key]["function"][func]["limits"],
                                   memMax=io[key]["function"][func]["membership_max"]) for xi in x]
                ax[i,j].plot(x,y,color=colors[k],label=func)
                ax[i,j].text(io[key]["value"],
                             io[key]["function"][func]["membership"], 
                             str(round(io[key]["function"][func]["membership"],2)),
                             color="w",fontweight="bold",
                             ha="center", va="center", bbox=dict(facecolor=colors[k],edgecolor='none',pad=1.0))
                if j > 0:
                    y = [fuzzy.memfunc(inpt  =xi,
                                       func  =io[key]["function"][func]["type"],
                                       limits=io[key]["function"][func]["limits"],
                                       memMax=io[key]["function"][func]["membership"]) for xi in x]
                    ax[i,j].fill_between(x,y,color=colors[k],alpha=0.4)
            leg = ax[i,j].legend(loc="upper center", ncol=len(io[key]["function"].keys()),handletextpad=0.,frameon=True)
            for c,l,t in zip(colors,leg.legendHandles,leg.get_texts()):t.set_color(c);t.set_ha('left');t.set_fontweight('bold');l.set_linewidth(0)
            ax[i,j].set_ylim(-0.1,1.55)
            ax[i,j].set_yticks([0,.2,.4,.6,.8,1])
            ax[i,j].plot([io[key]["value"],io[key]["value"]],[0,1],"w--",linewidth=1)
            ax[i,j].text(io[key]["value"], 1.08, str(round(io[key]["value"],2)),fontsize=11,ha="center") 
            title = "INPUT"
            if j > 0: title = "OUTPUT" 
            ax[i,j].set_title(title+": "+key)
            
    # Simulate multiple scenarios
    io_links = {}
    for rule in fuzzy.rules.keys():
        for otpt in fuzzy.rules[rule]["output"].keys():
            if otpt not in io_links.keys():io_links[otpt] = []
            for inpt in fuzzy.rules[rule]["input"].keys():
                if inpt not in io_links[otpt]: io_links[otpt].append(inpt)
    for i,otpt in enumerate(io_links.keys()):
        x = {}
        for inpt,N in zip(io_links[otpt][:2],[50,10]): x[inpt] = fuzzy.gen_xrange(fuzzy.input[inpt],N)
  
        if len(x.keys()) == 1:
            inpt = list(x.keys())[0]
            y = []
            for xi in x[inpt]:
                fuzzy.input[inpt]["value"] = xi
                fuzzy.gen_output()
                y.append(fuzzy.output[otpt]["value"])
            ax[i,2].plot(x[inpt],y,"w",drawstyle="steps-post")
            ax[i,2].set_ylabel("OUTPUT: "+otpt)
            ax[i,2].set_xlabel("INPUT: " +inpt)
        else:
            low_len_key = [(key, len(x[key])) for key in x.keys()]
            low_len_key.sort(key=lambda x: x[1])
            low_len_key, high_len_key  = low_len_key[0][0],low_len_key[1][0]
            
            colors = sns.color_palette("Spectral",len(x[low_len_key]))[::-1]
            y_prev = []
            for k,xi in enumerate(x[low_len_key]):
                fuzzy.input[low_len_key]["value"] = xi
                y = []
                for xii in x[high_len_key]:
                    fuzzy.input[high_len_key]["value"] = xii
                    fuzzy.gen_output()
                    y.append(fuzzy.output[otpt]["value"])
                if k==0 or sum([(abs(yi/np.std(y)-y_previ/np.std(y_prev))) for yi,y_previ in zip(y,y_prev)]) > 1e-10:
                    ax[i,2].plot(x[high_len_key],y,color=colors[k],label=str(round(xi,2)),drawstyle="steps-post")
                    ax[i,2].text(x[high_len_key][int(len(y)*.5)],y[int(len(y)*.5)],str(round(xi,2)),
                                 color=colors[k],fontweight="bold",
                                 ha="center", va="center", bbox=dict(facecolor="k",edgecolor='none',pad=1.0,alpha=1))
                y_prev = y
            ax[i,2].set_ylabel("OUTPUT: "+otpt)
            ax[i,2].set_xlabel("INPUT: " +high_len_key)
            leg = ax[i,2].legend(loc="upper right", 
                                 frameon = False, 
                                 title = "INPUT: " + low_len_key, 
                                 bbox_to_anchor=(1.4, 1.),
                                 ncol=int(len(ax[i,2].lines)>5)+1,
                                 handletextpad=0.)
            for l,t in zip(leg.legendHandles,leg.get_texts()):
                t.set_color(l.get_color())
                t.set_ha('center')
                t.set_fontweight('bold')
                l.set_linewidth(0)
            leg.get_title().set_position((-15, 0))
            leg._legend_box.align = "left"
    
    # Clean empty plots
    for i in range(ax.shape[0]):
        for j in range(ax.shape[1]):
            if len(ax[i,j].lines) == 0:
                fig.delaxes(ax[i,j])

if __name__ == "__main__":
    import os
    import sys
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from controllers.fuzzy import fuzzy

    if 1:
        # inputs
        l_c_rc = 20.
        y_c_rc = 26.
        h_c_rc = 27.
        
        l_c_rt = 0.4
        y_c_rt = 0.75
        h_c_rt = 0.8
        
        l_c_cc = 52.
        y_c_cc = 53.
        h_c_cc = 58.
        
        # outputs
        l_f_rc = 50
        h_f_rc = 120
        dr_f_rc_m = (h_f_rc-l_f_rc)/10
        
        l_v_s1 = 5
        h_v_s1 = 15
        dr_v_s1_m = (h_v_s1-l_v_s1)/10
        
        l_f_cc = 3
        h_f_cc = 10
        dr_f_cc_m = (h_f_cc-l_f_cc)/10

    fuzzy_c = fuzzy()

    # inputs
    p = [0.,.1,.3,.5,.7,.9,1.]
    fuzzy_c.input["y_c_rt"]   = {"value"   : y_c_rt, 
                                 "function":{"low"     :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[l_c_rt+(h_c_rt-l_c_rt)*p[0], l_c_rt+(h_c_rt-l_c_rt)*p[0], l_c_rt+(h_c_rt-l_c_rt)*p[2]]},
                                             "good"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[l_c_rt+(h_c_rt-l_c_rt)*p[1], l_c_rt+(h_c_rt-l_c_rt)*p[3], l_c_rt+(h_c_rt-l_c_rt)*p[5]]},
                                             "high"    :{"membership":0, "membership_max":1, "type":"rtrap", "limits":[l_c_rt+(h_c_rt-l_c_rt)*p[4], l_c_rt+(h_c_rt-l_c_rt)*p[6], l_c_rt+(h_c_rt-l_c_rt)*p[6]]}}}

    fuzzy_c.input["y_c_rc"]   = {"value"    : y_c_rc, 
                                 "function":{"low"     :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[l_c_rc+(h_c_rc-l_c_rc)*p[0], l_c_rc+(h_c_rc-l_c_rc)*p[0], l_c_rc+(h_c_rc-l_c_rc)*p[2]]},
                                             "good"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[l_c_rc+(h_c_rc-l_c_rc)*p[1], l_c_rc+(h_c_rc-l_c_rc)*p[3], l_c_rc+(h_c_rc-l_c_rc)*p[5]]},
                                             "high"    :{"membership":0, "membership_max":1, "type":"rtrap", "limits":[l_c_rc+(h_c_rc-l_c_rc)*p[4], l_c_rc+(h_c_rc-l_c_rc)*p[6], l_c_rc+(h_c_rc-l_c_rc)*p[6]]}}}

    fuzzy_c.input["y_c_cc"]   = {"value"    : y_c_cc, 
                                 "function":{"low"     :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[l_c_cc+(h_c_cc-l_c_cc)*p[0], l_c_cc+(h_c_cc-l_c_cc)*p[0], l_c_cc+(h_c_cc-l_c_cc)*p[2]]},
                                             "good"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[l_c_cc+(h_c_cc-l_c_cc)*p[1], l_c_cc+(h_c_cc-l_c_cc)*p[3], l_c_cc+(h_c_cc-l_c_cc)*p[5]]},
                                             "high"    :{"membership":0, "membership_max":1, "type":"rtrap", "limits":[l_c_cc+(h_c_cc-l_c_cc)*p[4], l_c_cc+(h_c_cc-l_c_cc)*p[6], l_c_cc+(h_c_cc-l_c_cc)*p[6]]}}}

    # outputs    
    fuzzy_c.output["dr_f_rc"] = {"value"   : 0, 
                                 "function":{"reduce"  :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[-dr_f_rc_m*2.09, -dr_f_rc_m, dr_f_rc_m     ]},
                                             "none"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[-dr_f_rc_m,       0,         dr_f_rc_m     ]},
                                             "increase":{"membership":0, "membership_max":1, "type":"rtrap", "limits":[-dr_f_rc_m,       dr_f_rc_m, dr_f_rc_m*2.09]}}}                                       

    fuzzy_c.output["dr_v_s1"] = {"value"   : 0, 
                                 "function":{"reduce"  :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[-dr_v_s1_m*2.09, -dr_v_s1_m, dr_v_s1_m     ]},
                                             "none"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[-dr_v_s1_m,       0,         dr_v_s1_m     ]},
                                             "increase":{"membership":0, "membership_max":1, "type":"rtrap", "limits":[-dr_v_s1_m,       dr_v_s1_m, dr_v_s1_m*2.09]}}}                                          
                           
    fuzzy_c.output["dr_f_cc"] = {"value"   : 0, 
                                 "function":{"reduce"  :{"membership":0, "membership_max":1, "type":"ltrap", "limits":[-dr_f_cc_m*2.09, -dr_f_cc_m, dr_f_cc_m     ]},
                                             "none"    :{"membership":0, "membership_max":1, "type":"tria" , "limits":[-dr_f_cc_m,       0,         dr_f_cc_m     ]},
                                             "increase":{"membership":0, "membership_max":1, "type":"rtrap", "limits":[-dr_f_cc_m,       dr_f_cc_m, dr_f_cc_m*2.09]}}}                               

    # rules
    #-----------------------------------------------------------
    fuzzy_c.rules["f_rc1"] = {"input" :{"y_c_rt" :"high", 
                                        "y_c_rc" :"high"},
                              "output":{"dr_f_rc":"increase"}}

    fuzzy_c.rules["f_rc2"] = {"input" :{"y_c_rt" :"low", 
                                        "y_c_rc" :"low"},
                              "output":{"dr_f_rc":"reduce"}}

    fuzzy_c.rules["f_rc3"] = {"input" :{"y_c_rt" :"good",
                                        "y_c_rc" :"good"},
                              "output":{"dr_f_rc":"none"}}
    #-----------------------------------------------------------
    fuzzy_c.rules["v_rt1"] = {"input" :{"y_c_rt" :"high"},
                              "output":{"dr_v_s1":"increase"}}
    
    fuzzy_c.rules["v_rt2"] = {"input" :{"y_c_rt" :"low"},
                              "output":{"dr_v_s1":"reduce"}}

    fuzzy_c.rules["v_rt3"] = {"input" :{"y_c_rt" :"good"},
                              "output":{"dr_v_s1":"none"}}
    #-----------------------------------------------------------
    fuzzy_c.rules["f_cc1"] = {"input" :{"y_c_cc" :"high"},
                              "output":{"dr_f_cc":"increase"}}
    
    fuzzy_c.rules["f_cc2"] = {"input" :{"y_c_cc" :"low"},
                              "output":{"dr_f_cc":"reduce"}}

    fuzzy_c.rules["f_cc3"] = {"input" :{"y_c_cc" :"good"},
                              "output":{"dr_f_cc":"none"}}
    #-----------------------------------------------------------
    view_fuzzy_design(fuzzy_c)
    plt.show()